#!/usr/bin/env python

"""
Optimise code produced by the AST translator.

Copyright (C) 2007, 2008, 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.common import *
from micropython.data import *
from micropython.rsvp import *

class Optimiser:

    "A code optimiser."

    supported_optimisations = [
        # Code generation optimisations:
        "constant_storage", "constant_accessor", "known_target", "self_access",
        "temp_storage", "no_operations", "unused_results",
        "unused_handlers", "accesses_by_usage",
        # Inspection optimisations:
        "unused_objects"
        ]

    def __init__(self, translation, optimisations=None):

        """
        Provide for the given 'translation' an optimiser for the desired
        'optimisations'. See the 'supported_optimisations' attribute of this
        class for permitted values.
        """

        self.translation = translation
        self.optimisations = set(optimisations or [])

        # The current "active" instruction.
        # As a rule, this will become the last instruction, but some
        # control-flow operations will flush the "active" instruction.

        self.active = None

        # Information about instructions providing the active values for
        # registers.

        self.saved_value_op = {}
        self.saved_value_block = {}
        self.saved_value_pos = {}

        # Sets of instructions providing the active value for each register.

        self.active_values = {}

    def get_attribute_store_instructions(self):

        """
        Return an appropriate set of instructions available when storing
        attributes.
        """

        ca = self.should_optimise_accesses_by_attribute_usage()

        return (
            ca and StoreAddress or None,        # plain assignment
            ca and StoreAddressContext or None, # assignment via class
            ca and StoreAddressContext or None, # assignment via class (never via instance)
            StoreAttr,                          # assignment via instance
            StoreAttrIndex,                     # special assignment via instance
            None
            )

    def reset(self, block=None, blocks=None):

        """
        Reset the optimiser for the given 'block' (or the current block),
        clearing the active value instructions if no 'blocks' are given, or
        collecting the active instructions from each of the blocks otherwise.
        """

        self.store_active_values()
        self.clear_active()

        # Make a new collection of instructions for a new block.

        if block:
            self.active_values = {}
            block.set_active_values(self.active_values)

        # Otherwise, clear the collection for an existing block.

        else:
            self.active_values.clear()

        if blocks:
            for b in blocks:
                self.active_values.update(b.get_active_values())

    def set_new(self, register, op):

        "For 'register', set the latest 'op' as the active instruction."

        self.set_active(op)
        self.set_active_value(register, op)

    def set_active_value(self, register, op):

        "For 'register', set 'op' as the value-providing active instruction."

        # Since the current value provider may eventually be used, it needs to
        # store its output if appropriate.

        self.store_active_value(register)
        self.active_values[register] = set([op])

    def get_active_value(self, register):

        """
        Get any single active value for the given 'register' or None if none or
        more than one exist.
        """

        if self.active_values.has_key(register) and \
            len(self.active_values[register]) == 1:

            return list(self.active_values[register])[0]

        else:
            return None

    def remove_active_value(self, register):

        """
        Remove the active instruction providing 'register' with its value from
        the generated code, if appropriate, and return the active instruction.
        """

        if self.active_values.has_key(register) and \
            self.active in self.active_values[register]:

            removed = self.translation.remove_op()
            self.active_values[register].remove(removed)
            return removed

        else:
            return None

    def set_target(self, register, target):

        """
        Set the target of the active instruction involving 'register' to
        'target'.
        """

        if self.active_values.has_key(register):
            for expr in self.active_values[register]:
                expr.target = target

            # Transfer the active instructions to their new target.

            if not self.active_values.has_key(target):
                self.active_values[target] = self.active_values[register]
            else:
                self.active_values[target].update(self.active_values[register])

            # Remove the association between the instructions and the specified
            # register.

            del self.active_values[register]

    def set_active(self, op):

        "Set the active instruction."

        self.active = op

    def clear_active(self):

        "Clear the active instructions."

        self.active = None

    # Permit the active value to be requested and restored.

    def request_active_value(self, register, temp, block, pos):

        """
        Request the current active value for 'register' so that if the value is
        changed, a temporary storage element or equivalent will be allocated.
        """

        # Cause any previously active value for the register to be saved.

        self.store_active_value(register)

        # Redefine the active value for the register.

        self.saved_value_op[register] = temp
        self.saved_value_block[register] = block
        self.saved_value_pos[register] = pos

    def store_active_values(self):

        "Store all active values."

        for register in self.active_values.keys():
            self.store_active_value(register)

    def store_active_value(self, register):

        "Store the requested active value for 'register'."

        if self.saved_value_op.has_key(register):

            # Cause an instruction to be inserted to store the active value for
            # the register, thus keeping it available for any subsequent usage.

            self.translation.set_temp(
                self.saved_value_op[register],      # cause the storage of the value
                self.saved_value_block[register],   # in the appropriate block
                self.saved_value_pos[register]      # at the appropriate location
                )

            # Discard the existing active value information.

            self.ignore_active_value(register)

    def ignore_active_value(self, register):

        "Ignore the active value in 'register'."

        if self.saved_value_op.has_key(register):
            del self.saved_value_op[register]
            del self.saved_value_block[register]
            del self.saved_value_pos[register]

    # Optimisation tests.

    def should_optimise_constant_storage(self):
        return "constant_storage" in self.optimisations

    def should_optimise_constant_accessor(self):
        return "constant_accessor" in self.optimisations

    def should_optimise_known_target(self):
        return "known_target" in self.optimisations

    def should_optimise_self_access(self):
        return "self_access" in self.optimisations

    def should_optimise_temp_storage(self):
        return "temp_storage" in self.optimisations

    def should_optimise_away_no_operations(self):
        return "no_operations" in self.optimisations

    def should_optimise_unused_results(self):
        return "unused_results" in self.optimisations

    def should_optimise_unused_handlers(self):
        return "unused_handlers" in self.optimisations

    def should_optimise_accesses_by_attribute_usage(self):
        return "accesses_by_usage" in self.optimisations

    # Simple tests.

    def is_constant_input(self, instruction):

        "Return whether 'instruction' provides a constant input."

        return isinstance(instruction, LoadAddress) and instruction.attr.assignments == 1 and \
            isinstance(instruction.attr.get_value(), Constant) or \
            isinstance(instruction, (LoadConst, LoadClass, LoadFunction))

    def is_constant_target(self, instruction):

        "Return whether 'instruction' provides a constant target."

        # NOTE: Removed StoreName, since this would then demand population of
        # NOTE: locals/frames.

        return isinstance(instruction, (StoreAddress, StoreAddressContext)) and \
            instruction.attr.is_constant() and \
            instruction.attr.is_strict_constant()

    def is_simple_input(self, instruction):

        """
        Return whether 'instruction' provides a simple input (typically a load
        instruction). A simple input is something which would be represented by
        a load operation from a CPU register or special memory location.
        """

        return isinstance(instruction, (LoadConst, LoadClass, LoadFunction,
            LoadName, LoadTemp, LoadAddress
            ))

    def is_resultant_no_operation(self, instruction, last_op=None):

        """
        Return whether 'instruction' merely stores its input where the input
        originally came from.
        """

        last_op = last_op or self.translation.last_op()
        return last_op and last_op.attr == instruction.attr and (
            isinstance(instruction, StoreTemp) and isinstance(last_op, LoadTemp) or
            isinstance(instruction, StoreAddress) and isinstance(last_op, LoadAddress) or
            last_op.source == instruction.target and (
                isinstance(instruction, LoadTemp) and isinstance(last_op, StoreTemp) or
                isinstance(instruction, LoadAddress) and isinstance(last_op, StoreAddress)
            ))

    # Convenience tests on outputs.

    def have_constant_target(self):

        "Return whether the active instruction provides a constant target."

        return self.is_constant_target(self.active)

    def have_constant_source(self, expr):

        "Return whether the active assignment source instruction is constant."

        return self.is_constant_input(expr)

    # Convenience tests on inputs.

    def have_constant_input(self):

        "Return whether the active instruction provides a constant input."

        return self.get_active_value("working") and self.is_constant_input(self.get_active_value("working"))

    def have_simple_input(self):

        "Return whether the active instruction provides a simple input."

        return self.get_active_value("working") and self.is_simple_input(self.get_active_value("working"))

    # Indicate whether the active instruction can be used in place of access
    # to a temporary variable retaining the result of the last instruction.

    have_temp_compatible_access = have_simple_input

    def have_self_input(self, unit):

        """
        Return whether the active instruction is a reference to self in the
        given 'unit'.
        """

        if not (isinstance(unit, Function) and unit.is_method()):
            return False

        expr = self.get_active_value("working")
        return expr and isinstance(expr, LoadName) and expr.attr.name == "self"

    def have_correct_self_for_target(self, context, unit):

        "Return whether the 'context' is compatible with the given 'unit'."

        if context is not None and self.have_self_input(unit):

            parent = unit.parent
            if parent is context or parent.has_subclass(context) or context.has_subclass(parent):
                return True

        return False

    def have_empty_handler(self, instruction):

        """
        Return whether the active instruction defined a handler for exceptions
        which is then discarded by the given 'instruction'.
        """

        return isinstance(self.translation.last_op(), PushHandler) and isinstance(instruction, PopHandler)

    # Optimisation methods. See the supported_optimisations class attribute.

    def optimise_constant_storage(self, expr):

        """
        Where the last operation stores a constant into a target which is also
        constant, indicate that both operations should be optimised away.
        """

        return self.should_optimise_constant_storage() and \
            self.have_constant_target() and \
            self.have_constant_source(expr)

    def optimise_constant_accessor(self):

        """
        Where the object whose attribute is being accessed is constant, provide
        information about the object and its full name.
        """

        if self.should_optimise_constant_accessor() and self.have_constant_input():
            value = self.get_active_value("working")

            # Get the details of the access.

            if isinstance(value.attr, Const):
                return value.attr, value.attr.value_type_name()
            else:
                target = value.attr.get_value()

                if target is None:
                    return None # no clearly defined target
                elif isinstance(target, Const):
                    return target, target.value_type_name()
                elif isinstance(target, Instance):
                    return None # skip production of optimised code
                else:
                    return target, target.full_name()

        else:
            return None

    def optimise_known_target(self):

        """
        Where the target of an invocation is known, provide information about it
        and its context. If a class is being invoked and the conditions are
        appropriate, get information about the specific initialiser.
        """

        if self.should_optimise_known_target() and self.have_constant_input():
            value = self.get_active_value("working")
            target = value.attr.get_value()
            context = value.attr.get_context()

            # Return target details only if this is clearly defined.

            if target is not None:
                return target, context

        return None

    def optimise_self_access(self, unit, attrname):

        """
        Return whether code in the given 'unit' is able to access the given
        'attrname' through the same position in all possible objects which might
        be accessed.
        """

        return self.should_optimise_self_access() and \
            self.have_self_input(unit) and not unit.is_relocated(attrname)

    def optimise_temp_storage(self):

        """
        Where the next operation would involve storing a value into temporary
        storage, record and remove any simple instruction which produced the
        value to be stored such that instead of subsequently accessing the
        temporary storage, that instruction is substituted.

        If no optimisation can be achieved, temporary storage is requested
        and the appropriate LoadTemp instruction is returned.

        Restriction: for use only in situations where the source of the
        temporary data will not be disturbed between its first access and its
        subsequent use.
        """

        if self.should_optimise_temp_storage():

            # Emitted instructions can be obtained.

            if self.have_temp_compatible_access():

                # Remove the active value contributor if possible.

                removed = self.remove_active_value("working")
                if removed is not None:

                    # Extend the lifetime of any temporary storage location.

                    self.translation.ensure_temp(removed)
                    return removed

                # Otherwise, just leave it in place, but return the instruction.

                else:
                    return self.get_active_value("working")

            # Or provisional temporary instructions.

            elif self.saved_value_op.has_key("working"):
                return self.saved_value_op["working"]

        return self.translation.get_temp()

    def optimise_away_no_operations(self, instruction, last_op=None):

        """
        Optimise away operations which just store their inputs in the place
        the inputs originally came from.
        """

        return self.should_optimise_away_no_operations() and \
            self.is_resultant_no_operation(instruction, last_op)

    def optimise_unused_results(self):

        "Discard results which will not be used."

        if self.should_optimise_unused_results() and self.have_simple_input():
            self.remove_active_value("working")

    def optimise_unused_handlers(self, instruction):

        "Discard handlers which will not be used."

        if self.should_optimise_unused_handlers() and self.have_empty_handler(instruction):
            self.translation.remove_op()
            return True
        else:
            return False

# vim: tabstop=4 expandtab shiftwidth=4
